#include <iostream>
#include <fstream>
#include <iomanip>
#include <ctime>
#include <cmath>
#include <sstream>
#include <csignal>
#include <vector>
#include <cstdlib>
#include <cerrno>
#include <climits>
#include <TROOT.h>
#include <TCanvas.h>
#include <TProfile2D.h>
#include <TF1.h>
#include <TH1.h>
#include "AsciiRoot.h"
#include "utils.h"
#include "Tracer.h"
#include "Hit.h"

#if HAVE_CONFIG_H
#ifdef G_OS_WIN32
#include "config_win.h"
#else
#include <config.h>
#endif
#endif

#ifndef HAS_SIGHANDLERT_T
#ifdef HAS_SIG_T
#define sighandler_t sig_t
#endif
#endif

struct AsciiRootPriv
{
    TH2F *histoPtr; //this is the peak and baseline histogram pointers
    TH2F *histoPtr2;
    TH1 *cluster1widebaselinehisto;
    TH1 *cluster2widebaselinehisto;
    TH1 *cluster3ormorewidebaselinehisto;
    TH1 *cluster1widepeakhisto;
    TH1 *cluster2widepeakhisto;
    TH1 *cluster3ormorewidepeakhisto;
    TH1 *imaxpeakhisto;
    TH1 *imaxbaselinehisto;
    TH1 *totalbaselinesig;
    TH1 *totalpeaksig;
    TH1 *Left1sthisto;
    TH1 *Left2ndhisto;
    TH1 *Right1sthisto;
    TH1 *Right2ndhisto;
    TH2D *gTempHisto;
    double gTempEventSum;
    double gTempSum;
    int gtimeslice1;
    int gtimeslice2;
    int gtimeslice3;
    int gtimeslice4;
    double gain1summation;
    double gain2summation;
};

bool _A_do_run = true;
void _A_got_intr(int)
{
    _A_do_run = false;
}

// decodes the header and returns a vector with the integers found
std::vector<int> decode_header(const std::string &h, AsciiRoot::XtraValues &xtra)
{
    std::vector<int> vout;
    std::istringstream istr(h);
    char *endptr, *str;
    char buf[256];
    long val;

    xtra.clear();

    while (istr)
    {
        istr.getline(buf, sizeof(buf), ';');
        if (!istr)
            break;

        errno = 0;
        val = strtol(buf, &endptr, 0);

        if ((errno == ERANGE && (val == LONG_MAX || val == LONG_MIN)) || (errno != 0 && val == 0))
        {
            std::string sval(buf), sout;
            sout = trim_str(sval);
            if (!sout.empty())
                xtra.push_back( sout );
        }
        else if ( endptr == buf || *endptr != '\0' )
        {
            std::string sval(buf), sout;
            sout = trim_str(sval);
            if (!sout.empty())
                xtra.push_back( sout );
        }
        else
        {
            vout.push_back(atoi(buf) );
        }
    }
    return vout;
}

AsciiRoot::AsciiRoot(const char *nam, const char *pedfile, const char *gainfile) :
    _nchan(max_nchan),_seedcut(10.), _neighcut(5.), _average_gain(1.), _version(2), _polarity(1),
    _priv(0)
{
    int i;
    for (i=0;i<max_nchan;i++)
    {
        _ped[i] = 0.;
        _gain[i] = 1.;
        _noise[i] = 1.;
        _data.data[i] = 0;
        _mask[i] = false;
    }
    init_priv_structure();

    if (nam)
        open(nam);

    if (pedfile)
        load_pedestals(pedfile);

    if (gainfile)
        load_gain(gainfile);

    // TODO: this loads a "fixed" name file. Be more general...
    // load_masking();
}
AsciiRoot::~AsciiRoot()
{
    if (ifile)
    {
        ifile->close();
    }
    if (_priv)
        delete _priv;
}
void AsciiRoot::init_priv_structure()
{
    if (_priv)
        delete _priv;

    _priv = new AsciiRootPriv;
    _priv->histoPtr=0; //this is the peak and baseline histogram pointers
    _priv->histoPtr2=0;
    _priv->cluster1widebaselinehisto=0;
    _priv->cluster2widebaselinehisto=0;
    _priv->cluster3ormorewidebaselinehisto=0;
    _priv->cluster1widepeakhisto=0;
    _priv->cluster2widepeakhisto=0;
    _priv->cluster3ormorewidepeakhisto=0;
    _priv->imaxpeakhisto=0;
    _priv->imaxbaselinehisto=0;
    _priv->totalbaselinesig=0;
    _priv->totalpeaksig=0;
    _priv->Left1sthisto=0;
    _priv->Left2ndhisto=0;
    _priv->Right1sthisto=0;
    _priv->Right2ndhisto=0;
    _priv->gTempHisto=0;
    _priv->gTempEventSum=0;
    _priv->gTempSum=0;
    _priv->gtimeslice1=5;
    _priv->gtimeslice2=15;
    _priv->gtimeslice3=80;
    _priv->gtimeslice4=100;
    _priv->gain1summation=0;
    _priv->gain2summation=0;
}

void AsciiRoot::open(const char *name)
{
    ifile = new std::ifstream(name);
    if (!(*ifile))
    {
        std::cout << "Could not open data file: " << name << std::endl;
        delete ifile;
        ifile = 0;
		return;
    }
    std::string header;
    unsigned int ic, lheader;
    char c;
    ifile->read((char *)&_t0, sizeof(time_t));
    //std::cout << "64b: " << ctime(&_t0) << std::endl;
    ifile->read((char *)&_type, sizeof(int));
    //std::cout << "type_ " << _type << std::endl;
    if ( _type > 15 )
    {
        ifile->seekg(0, std::ios::beg);
        ifile->read((char *)&_t0, sizeof(int));
        ifile->read((char *)&_type, sizeof(int));
        //std::cout << "32b: " << ctime(&_t0) << std::endl;
    }


    ifile->read((char *)&lheader, sizeof(unsigned int));
    for (ic=0; ic<80; ic++)
    {
        ifile->read(&c, sizeof(char));
        header.append(1, c);
    }
    header = trim_str(header);

    if (header[0]!='V' && header[0]!='v')
    {
        _version = 0;
    }
    else
    {
        _version = int(header[1]-'0');
        header = header.substr(5);
    }

    std::cout << "type: " << _type << " header: " << header << std::endl;
    std::vector<int> param = decode_header(header, _xtra);
    ifile->read((char *)_ped, max_nchan*sizeof(double));
    ifile->read((char *)_noise, max_nchan*sizeof(double));
    switch (_type)
    {
        case 1: // calibration
        case 2: // laser sync
            _npoints = param[0];
            _from = param[1];
            _to = param[2];
            _step = param[3];
            break;
        case 3: // laser run
        case 4: // source run
        case 5: // pedestal run
            if (param.empty())
                _nevts = 100000;
            else
                _nevts = param[0];
            _npoints = _from = _to = _step = 0;
            break;
    }
    data_start = ifile->tellg();
}

void AsciiRoot::rewind()
{
    if (ifile)
    {
        ifile->clear();
        ifile->seekg(data_start, std::ios::beg);
    }
}

void AsciiRoot::close()
{
    if (ifile)
    {
        ifile->close();
        delete ifile;
        ifile = 0;
    }
}

void AsciiRoot::reset_data()
{
    memset(&_data, 0, sizeof(_data));
}

int AsciiRoot::read_event()
{
    if (ifile)
    {
        unsigned int header, size, user=0, code=0;
        char *block_data=0;
        if (_version)
        {
            do
            {
                do
                {
                    ifile->read((char *)&header, sizeof(unsigned int));
                    if (ifile->bad() || ifile->eof())
                        return -1;

                    code = (header>>16) & 0xFFFF;
                } while ( code != 0xcafe );

                code = header & 0x0fff;
                user = header & 0x1000;
                switch (code)
                {
                    case NewFile:
                        ifile->read((char *)&size, sizeof(unsigned int));
                        block_data = new char[size];
                        ifile->read(block_data, size);
                        new_file(size, block_data);
                        break;
                    case StartOfRun:
                        ifile->read((char *)&size, sizeof(unsigned int));
                        block_data = new char[size];
                        ifile->read(block_data, size);
                        start_of_run(size, block_data);
                        break;
                    case DataBlock:
                        ifile->read((char *)&size, sizeof(unsigned int));
                        if (user)
                        {
                            reset_data();
                            block_data = new char[size];
                            ifile->read(block_data, size);
                            new_data_block(size, block_data);
                        }
                        else
                        {
                            if ( _version == 1 )
                            {
                                ifile->read((char *)&_data, sizeof(EventData));
                                for (int ii=0; ii<2; ii++)
                                    memset(_header[ii], 0, 16*sizeof(unsigned short));

                            }
                            else
                            {
                                ifile->read((char *)&_data.value, sizeof(double));
                                ifile->read((char *)&_data.time, sizeof(unsigned int));
                                ifile->read((char *)&_data.temp, sizeof(unsigned short));
                                for (int ii=0; ii<2; ii++)
                                {
                                    ifile->read((char *)_header[ii], 16*sizeof(unsigned short));
                                    ifile->read((char *)&_data.data[ii*128], 128*sizeof(unsigned short));
                                }
                            }
                        }

                        break;
                    case CheckPoint:
                        ifile->read((char *)&size, sizeof(unsigned int));
                        block_data = new char[size];
                        ifile->read(block_data, size);
                        check_point(size, block_data);
                        break;
                    case EndOfRun:
                        ifile->read((char *)&size, sizeof(unsigned int));
                        block_data = new char[size];
                        ifile->read(block_data, size);
                        end_of_run(size, block_data);
                        break;
                    default:
                        std::cout << "Unknown block data type: " << std::hex << header << " - " << code << std::dec << std::endl;
                }
                if (block_data)
                {
                    delete [] block_data;
                    block_data = 0;
                }

            } while ( code != DataBlock && !(ifile->bad() || ifile->eof()) );
        }
        else
        {
            ifile->read((char *)&_data, sizeof(EventData));
            for (int ii=0; ii<2; ii++)
                memset(_header[ii], 0, 16*sizeof(unsigned short));
        }

        if (ifile->eof())
        {
            std::cout << "End of file" << std::endl;
            return -1;
        }
        else if (ifile->bad())
        {
            std::cout << "Problems with data file" << std::endl;
            return -1;
        }
        else
        {
            //process_event();
            return 0;
        }
    }
    else
        return -1;
}

void AsciiRoot::set_data(int nchan, const unsigned short int *data)
{
    int i;
    _nchan = nchan;
    for (i=0;i<_nchan;i++)
        _data.data[i] = data[i];
}


double AsciiRoot::time() const
{
    unsigned short fpart = _data.time & 0xffff;
    short ipart = (_data.time & 0xffff0000)>>16;
    if (ipart<0)
        fpart *= -1;
    //double tt = 100.*(1. -(ipart + (fpart/65535.)));
    double tt = 100.0*(ipart + (fpart/65535.));
    return tt;
}

double AsciiRoot::temp() const
{
    return 0.12*_data.temp - 39.8;
}

double AsciiRoot::average_temp()
{
  double averagetemp=0;
  return averagetemp=_priv->gTempSum/_priv->gTempEventSum;
}

TH1 *AsciiRoot::show_pedestals()
{
    int ic;
    TH1 *hst = new TH1D("hPed","Pedestals",nchan(),-0.5, nchan()-0.5);
    hst->SetYTitle("ADCs");
    hst->SetXTitle("Channel no.");
    for (ic=0; ic<nchan(); ic++)
        hst->SetBinContent(ic+1, _ped[ic]);

    return hst;
}

TH1 *AsciiRoot::show_noise()
{
    int ic;
    TH1 *hst = new TH1D("hNoise","Noise",nchan(),-0.5, nchan()-0.5);
    if (gain()==1)
    {
        hst->SetYTitle("ADCs");
    }
    else
    {
        hst->SetYTitle("e^{-} ENC");
    }
    hst->SetXTitle("Channel no.");
    for (ic=0; ic<nchan(); ic++)
        hst->SetBinContent(ic+1, noise(ic));

    return hst;
}

void AsciiRoot::compute_pedestals_fast(int mxevts, double wped, double wnoise)
{
    int i, ievt;

    if (!ifile)
        return;

    if (mxevts<0)
        mxevts = 100000000;

    for (i=0;i<max_nchan;i++)
        _ped[i] = _noise[i] = 0.;

    std::ifstream::pos_type here = ifile->tellg();
    std::cout << "Computing fast pedestas..." << std::endl;
    for (ievt=0; read_event()==0 && ievt<mxevts; ievt++)
    {
        if (!(ievt%100))
        {
            std::cout << "\revent " << std::setw(10) << ievt << std::flush;
        }
        common_mode();
        for (i=0; i<nchan(); i++)
        {
            // TODO: figure out how to determine the chip number when
            //       Plugin::filter_event has been called
            int ichip = i/128;
            // IF noise is 0, set it arbitrarily to 1.
            if (_noise[i]==0.)
                _noise[i] = 1.;

            if (_ped[i]==0.)
            {
                // If pedestal is not yet computed we assume the current
                // channel value should not be too far
                _ped[i] = _data.data[i];
            }
            else
            {
                // Do the pedestal and noise correction
                double corr;
                double xs;

                _signal[i] = _data.data[i] - _ped[i];
                corr = _signal[i] * wped;

                xs = (_signal[i]-_cmmd[ichip])/_noise[i];
                if (corr > 1.)
                    corr = 1.;

                if (corr < -1)
                    corr = -1.;

                _ped[i] += corr;

                if (fabs(xs) < 3.)
                {
                    _noise[i] = _noise[i]*(1.0-wnoise) + xs*xs*wnoise;
                }
            }
        }
    }
    std::cout << "\nDone" << std::endl;
    rewind();
}

void AsciiRoot::set_timeslices(int timeslice1, int timeslice2, int timeslice3, int timeslice4)
{
  _priv->gtimeslice1=timeslice1;
  _priv->gtimeslice2=timeslice2;
  _priv->gtimeslice3=timeslice3;
  _priv->gtimeslice4=timeslice4;
}

TH2 *AsciiRoot::compute_pedestals(int mxevts, bool do_cmmd)
{
    if (!ifile)
        return 0;

    if (mxevts<0)
        mxevts = 100000000;

    int ievt, ichan;
    TH2 *hst = create_h2("hRaw","Raw data",nchan(), -0.5,nchan()-0.5, 256, -0.5,1023.5);
    TH2 *hsts = create_h2("hSig","Signal",nchan(), -0.5,nchan()-0.5,256, -127.5,127.5);


    std::ifstream::pos_type here = ifile->tellg();
    std::cout << "Computing pedestas..." << std::endl;
    for (ievt=0; read_event()==0 && ievt<mxevts; ievt++)
    {
        process_event(do_cmmd);
        for (ichan=0; ichan<nchan(); ichan++)
            // TODO: get right chip number in all situations (after calling set_data)
            hst->Fill(ichan, data(ichan)-get_cmmd(ichan/128));

        if (!(ievt%100))
        {
            std::cout << "\revent " << std::setw(10) << ievt << std::flush;
        }
    }
    std::cout << "\nDone" << std::endl;
    rewind();

    // TODO: _nchan can be updated in an event by event basis
    //       while here we are assuming that it is the same
    //       for all the events
    for (ichan=0; ichan<nchan(); ichan++)
    {
        TF1 *g = new TF1("g1", "gaus");
        TH1 *h1 = hst->ProjectionY("__hx__", ichan+1, ichan+1);
        g->SetParameters(h1->GetSumOfWeights(), h1->GetMean(), h1->GetRMS());
        g->SetRange(h1->GetMean()-2.5*h1->GetRMS(), h1->GetMean()+2.5*h1->GetRMS());
        h1->Fit("g1", "q0wr");
        _ped[ichan] = h1->GetFunction("g1")->GetParameter(1);
        _noise[ichan] = h1->GetFunction("g1")->GetParameter(2);
        delete h1;
        delete g;
    }

    rewind();
    for (ievt=0; read_event()==0 && ievt<mxevts; ievt++)
    {
        process_event(do_cmmd);
        for (ichan=0; ichan<nchan(); ichan++)
            hsts->Fill(ichan, signal(ichan));

        if (!(ievt%100))
        {
            std::cout << "\revent " << std::setw(10) << ievt << std::flush;
        }
    }
    std::cout << "\nDone" << std::endl;
    rewind();

    return hst;
}

void AsciiRoot::show_peakclusters()
{
  TVirtualPad *pad;
  TCanvas *henrycanvas = new TCanvas("Henrycanvas", "Henrycanvas", 300, 400);
  TH1 *peakclusters = create_h1("peakclusters", "Peak Cluster Density", 10, -0.5, 9.5);
  for (int i=0; i<20; i++)
    {
	 if (_hist1clusterwidth[i]!=0)
	    {
	      peakclusters->Fill(i,_hist1clusterwidth[i]);
	    }
    }
  pad = henrycanvas->cd(1);
  std::cout << "Peak cluster Histogram filled " << std::endl;
  peakclusters->SetXTitle("Cluster width");
  peakclusters->SetYTitle("Frequency");
  peakclusters->Draw();
  henrycanvas->Update();
}

void AsciiRoot::show_baselineclusters()
{
  TVirtualPad *pad;
  TCanvas *henrycanvas2 = new TCanvas("Henrycanvas2", "Henrycanvas2", 300, 400);
  TH1F *baselineclusters = new TH1F("baselineclusters", "Baseline Cluster Density", 10, -0.5, 9.5);
  for (int i=0; i<20; i++)
    {
    	  if (_hist2clusterwidth[i]!=0)
	    {
	      baselineclusters->Fill(i,_hist2clusterwidth[i]);
	      //std::cout << "Added one to the width " << i << std::endl;
	    }
    }
  pad = henrycanvas2->cd(1);
  std::cout << "Baseline cluster Histogram filled " << std::endl;
  baselineclusters->SetYTitle("Frequency");
  baselineclusters->SetXTitle("Cluster Width");
  baselineclusters->Draw();
  henrycanvas2->Update();
}

void AsciiRoot::find_clusters(int ichip)
{
    // TODO: figure out how to determine the chip number in
    //       all the situations
    int i, imax=-1, left, right;
    double mxsig=-1.e20, sg, val;
    bool used[256];
    int chan0=0;
    int chan1=256;

    if (!_priv->histoPtr)
        generate_clusterhistos();

    _priv->gTempSum+= temp();
    _priv->gTempEventSum += 1;
    if (_priv->gTempHisto)
        _priv->gTempHisto->Fill(_priv->gTempEventSum, temp());

    if (ichip>=0 && ichip<2)
    {
        chan0 = ichip*128;
        chan1 = (ichip+1)*128;
    }


    for (i=0;i<max_nchan;i++)
    {
        used[i]= _mask[i] ? true : false;
    }
    clear();


    while (true)
    {
        /*
         * Find the highest
         */
        imax = -1;
        for (i=chan0; i<chan1; i++)
        {
            if (used[i] || _signal[i]*polarity()<0.)
                continue;

            if ( polarity()*sn(i) > _seedcut)
            {
                val = fabs(signal(i));
                if (mxsig<val)
                {
                    mxsig = val;
                    imax = i;
                }
            }
        }

        if (imax<0 || imax >= max_nchan )
            break;

        sg = signal(imax);
        used[imax]=true;
        // Now look at neighbours
        // first to the left
        left = imax;
        for (i=imax-1;i>=0;i--)
        {
            if ( used[i] || _signal[i]*polarity()<0.)
                break;

            if ( fabs(sn(i)) > _neighcut )
            {
                used[i] = true;
                sg += signal(i);
                left = i;
            }
            else
                // TODO: this needs to be removed
                // The idea is to merge to clusters that have only one strip in between
                // In the laser runs this is a consequences of reflections...
            {
                int j = i-1;
                if (j>=0  )
                {
                    if ( fabs(sn(j)) > _neighcut )
                        continue;
                }
                break;
            }
        }
        // now to the right
        right = imax;
        for (i=imax+1;i<nchan();i++)
        {
            if ( used[i] || _signal[i]*polarity()<0.)
                break;
            if ( fabs(sn(i))>_neighcut )
            {
                used[i] = true;
                sg += signal(i);
                right = i;
            }
            else
                // TODO: this needs to be removed
                // The idea is to merge to clusters that hanve only one strip in between
                // In the laser runs this is a consequences of reflections...
            {
                int j = i+1;
                if (j<nchan())
                {
                    if ( fabs(sn(j)) > _neighcut )
                        continue;
                }
                break;
            }
        }
        add_hit(Hit(imax, left, right, sg));

        int henrywidth=0; //h
        _hist1clusterwidth[0]=0;
        if (imax!=-1)
        {
            henrywidth = right-left+1; //h
            // if (henrywidth >=15) henrywidth=15;
            // _clusterwidths[henrywidth] += 1;
        }
        if (henrywidth>20)
            henrywidth = 20;

        if (time()>= _priv->gtimeslice1 && time()<=_priv->gtimeslice2)
        {
            //			    double henrydata;
            //			    henrydata = (_data.data[imax]-_ped[imax]-_cmmd[imax/128])/sg;
            //			    std::cout << "********************************************************************************" << std::endl;
            //			    std::cout << "Cluster width " << right-left+1 << " fraction signal in main channel  " << henrydata << " starting channel " << imax << std::endl;
            //			    std::cout << "Total signal value is " << sg << std::endl;
            //			    for (i=left;i<=right;i++)
            //			      {
            //				std::cout <<"Chan " << i <<  " data " << _data.data[i] << ", pedestal " << _ped[i] << ", cmmd " << _cmmd[imax/128] << " signal " << _signal[i] << " s/n " << _sn[i] <<  std::endl;
            //			      }
            //			    std::cout << "*********************************************************************************" << std::endl;
            _hist1clusterwidth[henrywidth] += 1;
            //	std::cout << "henrywidth is " << henrywidth << " and signal is " << sg << std::endl;
            //	std::cout<<"_hist1clusterwidth[henrywidth] is "<<_hist1clusterwidth[henrywidth]<<st ::endl;
            _priv->imaxpeakhisto->Fill(imax);
            _priv->histoPtr->Fill(henrywidth, sg);
            if (henrywidth==1)
            {
                _priv->cluster1widepeakhisto->Fill(sg);
            }
            if (henrywidth==2)
            {
                _priv->cluster2widepeakhisto->Fill(sg);
            }
            if (henrywidth>=3)
            {
                _priv->cluster3ormorewidepeakhisto->Fill(sg);
            }
            _priv->totalpeaksig->Fill(sg);
            if (left<imax)
            {
                _priv->Left1sthisto->Fill(signal(imax-1));
                if (left<imax-1)
                {
                    _priv->Left2ndhisto->Fill(signal(imax-2));
                }
            }
            if (right>imax)
            {
                _priv->Right1sthisto->Fill(signal(imax+1));
                if (right>imax+1)
                {
                    _priv->Right2ndhisto->Fill(signal(imax+2));
                }
            }
        }
        if (time()>=_priv->gtimeslice3 && time() <=_priv->gtimeslice4)
        {
            _hist2clusterwidth[henrywidth] += 1;
            _priv->imaxbaselinehisto->Fill(imax);
            _priv->histoPtr2->Fill(henrywidth, sg);
            if (henrywidth==1)
            {
                _priv->cluster1widebaselinehisto->Fill(sg);
            }
            if (henrywidth==2)
            {
                _priv->cluster2widebaselinehisto->Fill(sg);
            }
            if (henrywidth>=3)
            {
                _priv->cluster3ormorewidebaselinehisto->Fill(sg);
            }
            _priv->totalbaselinesig->Fill(sg);
        }
    }
}

void AsciiRoot::save_pedestals(const char *fnam)
{
    std::ofstream ofile(fnam);
    if (!ofile)
    {
        std::cout << "Could not open " << fnam << " to save pedestals." << std::endl;
        return;
    }

    // TODO: _nchan can be updated in an event by event basis
    //       while here we are assuming that it is the same
    //       for all the events
    int i;
    for (i=0; i<nchan(); i++)
    {
        ofile << _ped[i] << "\t" << _noise[i] << "\n";
    }
    ofile.close();
}

void AsciiRoot::load_pedestals(const char *fnam)
{
    std::ifstream ifile(fnam);
    if (!ifile)
    {
        std::cout << "Could not open " << fnam << " to load pedestals." << std::endl;
        return;
    }
    int i;
    for (i=0; i<max_nchan; i++)
    {
        if (ifile.eof())
            break;

        ifile >> _ped[i] >> std::ws >> _noise[i] >> std::ws;
                 _mask[i] = (_noise[1]>20. || _noise[i]<=0.);
    }
    ifile.close();
    TCanvas *pedcanvas = new TCanvas("Pedcanvas", "Pedestal Values", 600, 400);
    TH1F *pedestalhisto = new TH1F("pedestalhisto", "Pedestal Values", 256, -0.5, 255.5);
    for (i=0; i<256; i++)
      {
	pedestalhisto->Fill(i, _ped[i]);
      }
    pedcanvas->cd(1);
    pedestalhisto->Draw();
}

void AsciiRoot::load_masking()
{
    std::ifstream ifile("masked.txt");
    if (!ifile)
    {
        std::cout << "Could not open masked.txt. " << std::endl;
        return;
    }
    int val;
    for (int i=0; i<500; i++)
    {
        ifile >> val >> std::ws;
        if (ifile.eof())
            break;
        if (val>255)
        {
            std::cout << "A value is greater than 255, causing an overflow crash. Please check the text file again. It has been set to 1 for continuation purposes. " << std::endl;
            val = 1;
        }
        _mask[val] = true;
    }
}

void AsciiRoot::load_gain(const char *fnam)
{
    std::ifstream ifile(fnam);
    if (!ifile)
    {
        std::cout << "Could not open " << fnam << " to load the gain." << std::endl;
        return;
    }
    int i;
    int ichan;
    double val, xn, xm;
    xn=xm=0.;
    for (i=0; i<max_nchan; i++)
    {
        ifile >> ichan >> std::ws;
        if (ifile.eof())
            break;
        ifile >> val;
        if (ifile.eof())
            break;

        xn++;

        // TODO: Henry hardcodes the gain to one. Why ?
        xm += val;
        _gain[ichan] = val;

        ifile >> std::ws;
        if (ifile.eof())
            break;
    }
    if (xn>0)
    {
        _average_gain = xm/xn;
    }
    ifile.close();
}

void AsciiRoot::process_event(bool do_cmmd)
{
    int i;
    for (i=0; i<nchan(); i++)
    {
        _signal[i] = _data.data[i]-_ped[i];
        _sn[i] = _noise[i]>1. && !_mask[i] ? _signal[i]/_noise[i] : 0.;
    }
    if (do_cmmd)
    {
        int ichip=-1;
        common_mode();

        for (i=0; i<nchan(); i++)
        {
            // TODO: figure out the right chip number
            if (!(i%128))
                ichip ++;

            _signal[i] = _data.data[i]-_ped[i] - _cmmd[ichip];
            _sn[i] = (_noise[i] >1. && !_mask[i] ? _signal[i]/_noise[i] : 0.);
        }
    }
}
void AsciiRoot::common_mode()
{
    int ip, i, j;


    for (j=0; j<2; j++)
    {
        int offs = 128*j;
        int sstop = 128*(j+1);
        double mean, sm, xn, xx, xs, xm, tmp;
        bool use_it;
        mean = sm = 0.;
        //std::cout << "Chip " << j << std::endl;
        for (ip=0;ip<3;ip++)
        {
            xn = xs = xm = 0.;
            for (i=offs; i<sstop; i++)
            {
                if (_mask[i])
                    continue;

                use_it = true;
                xx = data(i) - _ped[i];
                if (ip)
                {
                    tmp = fabs((xx-mean)/sm);
                    use_it = (tmp<2.5);
                }
                if (use_it)
                {
                    xn++;
                    xm += xx;
                    xs += xx * xx;
                }
            }
            if (xn>0.)
            {
                mean = xm / xn;
                sm = sqrt( xs/xn - mean*mean);
            }
          //  std::cout << "...iter " << ip << ": xm " << mean << " xs: " << sm << std::endl;
        }
        _cmmd[j] = mean;
        _cnoise[j] = sm;
    }
}

void AsciiRoot::generate_clusterhistos()
{
    // First we generate the histograms and then assign pointers
    TH2F *clusterpeaksignaldist = new TH2F("clusterpeaksignaldist", "Peak Clusters Signal Distribution", 10, -0.5, 9.5, 100, -1599.5, 0.5);
    TH2F *clusterbaselinesignaldist = new TH2F("clusterbaselinesignaldist", "Baseline Cluster Signal Distribution", 10, -0.5, 9.5, 100, -1599.5, 0.5);
    _priv->histoPtr = clusterpeaksignaldist;
    _priv->histoPtr2 = clusterbaselinesignaldist;
    TH1 *baselineone = create_h1("baselineone", "Baseline Clusters one strip wide", 80, -799.5, -0.5);
    TH1 *baselinetwo = create_h1("baselinetwo", "Baseline Clusters two strips wide", 80, -799.5, -0.5);
    TH1 *baselinethree = create_h1("baselinethree", "Baseline Clusters three or more wide", 80, -799.5, -0.5);
    _priv->cluster1widebaselinehisto = baselineone;
    _priv->cluster2widebaselinehisto = baselinetwo;
    _priv->cluster3ormorewidebaselinehisto = baselinethree;
    TH1 *peakone = create_h1("peakone", "Peak Clusters one strip wide", 80, -799.5, -0.5);
    TH1 *peaktwo = create_h1("peaktwo", "Peak Clusters two strips wide", 80, -799.5, -0.5);
    TH1 *peakthree = create_h1("peakthree", "Peak Clusters three or more wide", 80, -799.5, -0.5);
    _priv->cluster1widepeakhisto = peakone;
    _priv->cluster2widepeakhisto = peaktwo;
    _priv->cluster3ormorewidepeakhisto = peakthree;
    TH1 *peakimax = create_h1("peakimax", "Peak Clusters Frequency Channel Occurance", 256, -0.5, 255.5);
    TH1 *baselineimax = create_h1("baselineimax", "Baseline Clusters Frequency Channel Occurance", 256, -0.5, 255.5);
    _priv->imaxbaselinehisto = baselineimax;
    _priv->imaxpeakhisto = peakimax;
    TH1 *baselinesignal = create_h1("baselinesignal", "Baseline Signal (total)", 50, -400.5, -0.5);
    _priv->totalbaselinesig = baselinesignal;
    TH1 *peaksignal = create_h1("peaksignal", "Peak Signal (total)", 80, -799.5, -0.5);
    _priv->totalpeaksig = peaksignal;
    TH1 *Leftonehisto = create_h1("leftonehisto", "First neighbour to the left", 50, -199.5, 0.5);
    _priv->Left1sthisto=Leftonehisto;
    TH1 *Lefttwohisto = create_h1("lefttwohisto", "Second neighbour to the left", 50, -199.5, 0.5);
    _priv->Left2ndhisto=Lefttwohisto;
    TH1 *Rightonehisto = create_h1("rightonehisto", "First neighbour to the right", 50, -199.5, 0.5);
    TH1 *Righttwohisto = create_h1("righttwohisto", "Second neighbour to the right", 50, -199.5, 0.5);
    _priv->Right1sthisto=Rightonehisto;
    _priv->Right2ndhisto=Righttwohisto;
    TH2D *temphisto = new TH2D("temp histo", "Temperature by Event histogram", 100000, 0.5, 100000.5, 30, -25, -10);
    _priv->gTempHisto = temphisto;
    std::cout << "Histograms generated " << std::endl;
}

void AsciiRoot::draw_histograms()
{
	TVirtualPad *pad;
	TCanvas *distributioncanvas = new TCanvas("distributioncanvas", "distributioncanvas", 600, 300);
	TCanvas *baselinehistogramcanvas = new TCanvas("baselinedistributioncanvas", "Baseline Distribution Canvas", 600, 400);
	TCanvas *peakhistogramcanvas = new TCanvas("peakdistributioncanvas", "Peak Distribution Canvas", 600, 400);
	TCanvas *imaxhistos = new TCanvas("imaxhistos", "Imax Histograms", 600, 400);
	TCanvas *neighbourhistos = new TCanvas("neighbourhistos", "Neighbour Histograms", 600, 400);
	TCanvas *tempcanvas = new TCanvas("Temperature Canvas", "Temperature Canvas", 600, 400);
        pad=tempcanvas->cd(1);
	pad->SetGrid(1,1);
	_priv->gTempHisto->Draw();
	neighbourhistos->Divide(2,2);
	pad=neighbourhistos->cd(1);
	pad->SetGrid(1,1);
	_priv->Left1sthisto->Draw();
	pad=neighbourhistos->cd(2);
	pad->SetGrid(1,1);
	_priv->Left2ndhisto->Draw();
	pad=neighbourhistos->cd(3);
	pad->SetGrid(1,1);
	_priv->Right1sthisto->Draw();
	pad=neighbourhistos->cd(4);
	pad->SetGrid(1,1);
	_priv->Right2ndhisto->Draw();
	imaxhistos->Divide(2,1);
	pad=imaxhistos->cd(1);
	pad->SetGrid(1,1);
	_priv->imaxbaselinehisto->Draw();
	pad=imaxhistos->cd(2);
	pad->SetGrid(1,1);
	_priv->imaxpeakhisto->Draw();
	distributioncanvas->Divide(2,1);
	pad=distributioncanvas->cd(1);
	pad->SetGrid(1,1);
	_priv->histoPtr->Draw();
	pad=distributioncanvas->cd(2);
	pad->SetGrid(1,1);
	_priv->histoPtr2->Draw();
	baselinehistogramcanvas->Divide(2,2);
	pad=baselinehistogramcanvas->cd(1);
	pad->SetGrid(1,1);
	_priv->cluster1widebaselinehisto->Draw();
	pad=baselinehistogramcanvas->cd(2);
	pad->SetGrid(1,1);
	_priv->cluster2widebaselinehisto->Draw();
	pad=baselinehistogramcanvas->cd(3);
	pad->SetGrid(1,1);
	_priv->cluster3ormorewidebaselinehisto->Draw();
	pad=baselinehistogramcanvas->cd(4);
	pad->SetGrid(1,1);
	_priv->totalbaselinesig->Draw();
	peakhistogramcanvas->Divide(2,2);
	pad=peakhistogramcanvas->cd(1);
	pad->SetGrid(1,1);
	_priv->cluster1widepeakhisto->Draw();
	pad=peakhistogramcanvas->cd(2);
	pad->SetGrid(1,1);
	_priv->cluster2widepeakhisto->Draw();
	pad=peakhistogramcanvas->cd(3);
	pad->SetGrid(1,1);
	_priv->cluster3ormorewidepeakhisto->Draw();
	pad=peakhistogramcanvas->cd(4);
	pad->SetGrid(1,1);
	_priv->totalpeaksig->Draw();
	show_baselineclusters();
	show_peakclusters();
	std::cout << " Histograms drawn " << std::endl;
}


void AsciiRoot::average_values()
{
  int i=0;
  double sum_noise_1=0;
  double sum_noise_2=0;
  double sum_gain_1=0;
  double sum_gain_2=0;
  double average_noise_1=0;
  double average_noise_2=0;
  double average_gain_1=0;
  double average_gain_2=0;
  for (i=0; i<128; i++)
    {
      sum_noise_1 += _noise[i];
    }
  for (i=128; i<256; i++)
    {
      sum_noise_2 += _noise[i];
}
  sum_gain_1=_priv->gain1summation;
  sum_gain_2=_priv->gain2summation;
  average_noise_1 = sum_noise_1/128;
  average_noise_2 = sum_noise_2/128;
  average_gain_1 = sum_gain_1/128;
  average_gain_2 = sum_gain_2/128;
  std::cout << "The average noise on chip 1 is " << average_noise_1 << " and the average gain is " << average_gain_1 << " " << std::endl;
  std::cout << "The average noise on chip 2 is " << average_noise_2 << " and the average gain is " << average_gain_2 << " " << std::endl;
}


void AsciiRoot::add_gain(double gain, double ichan)
{
  if (ichan<=127)
    {
      _priv->gain1summation+=gain;
    }
  if (ichan>=128 && ichan <256)
    {
      _priv->gain2summation+=gain;
    }
}


void AsciiRoot::spy_data(bool with_signal, int nevt)
{
    TVirtualPad *pad;
    if (!ifile)
        return;

    sighandler_t old_handler = ::signal(SIGINT, _A_got_intr);
    _A_do_run = true;

    TCanvas *cnvs = (TCanvas *)gROOT->FindObject("cnvs");
    if (cnvs)
    {
        cnvs->Clear();
    }
    else
       cnvs = new TCanvas("cnvs","cnvs", 700, 800);

    cnvs->Divide(2,3);


    TH1 *hsignal = create_h1("hsignal","signal (ADC)",256, -0.5, 255.0);
    hsignal->SetXTitle("Channel");
    hsignal->SetYTitle("ADC");
    hsignal->SetMinimum(-300);
    hsignal->SetMaximum(300);

    TH1 *helec = create_h1("helec","signal (elec)", 256, -0.5, 255.5);
    helec->SetXTitle("Channel");
    helec->SetYTitle("electrons");
    helec->SetMinimum(-300/gain());
    helec->SetMaximum(300/gain());

    TH1 *hraw = create_h1("hraw","Raw Data (around 512.)",256, 0., 256.);
    hraw->SetXTitle("Channel");
    hraw->SetYTitle("ADC");
    hraw->SetMinimum(-300);
    hraw->SetMaximum(+300);

    TH1 *hrawc = create_h1("hrawc","Raw Data (no commd)",256, 0., 256.);
    hrawc->SetXTitle("Channel");
    hrawc->SetYTitle("ADC");
    hrawc->SetMinimum(-300);
    hrawc->SetMaximum(+300);


    TH1 *hcmmd[2];
    hcmmd[0] = create_h1("hcmmd0","Common mode (Chip 0)",50,-100.,100.);
    hcmmd[0]->SetXTitle("Common mode");
    hcmmd[1] = create_h1("hcmmd1","Common mode (Chip 1)",50,-100.,100.);
    hcmmd[1]->SetXTitle("Common mode");

    int ievt,jevt;
    for (ievt=jevt=0; read_event()==0 && _A_do_run && ievt<nevt;jevt++)
    {
        process_event();
        find_clusters();
        if ( with_signal && empty())
            continue;

        int i,ichip=-1;
        for (i=0; i<nchan(); i++)
        {
            // TODO: figure out chip number
            if (!(i%128))
                ichip++;

            hsignal->SetBinContent(i+1, _signal[i]);
            helec->SetBinContent(i+1, signal(i));
            hraw->SetBinContent(i+1,data(i)-512.);
            hrawc->SetBinContent(i+1, data(i)-_ped[i]);
            // TODO: why we draw the signal + common mode ?
            //       May be cause signal should be ~0...
            hcmmd[ichip]->Fill(_signal[i]+get_cmmd(ichip));
        }
        pad = cnvs->cd(1);
        pad->SetGrid(1,1);
        hsignal->Draw();
        pad = cnvs->cd(2);
        pad->SetGrid(1,1);
        helec->Draw();

        pad = cnvs->cd(3);
        pad->SetGrid(1,1);
        hraw->Draw();

        pad = cnvs->cd(4);
        pad->SetGrid(1,1);
        hrawc->Draw();

        pad = cnvs->cd(5);
        pad->SetGrid(1,1);
        hcmmd[0]->Draw();

        pad = cnvs->cd(6);
        pad->SetGrid(1,1);
        hcmmd[1]->Draw();

        std::cout << std::setiosflags(std::ios::fixed);
        std::cout << "*** Event " << jevt << " *****" << std::endl;
        std::cout << "Common Mode:" << std::endl
                  << "   Chip 0 " << std::setw(6) << std::setprecision(1) << get_cmmd(0) << " noise: " << get_cnoise(0)
                  << std::endl
                  << "   Chip 1 " << std::setw(6) << std::setprecision(1) << get_cmmd(1) << " noise: " << get_cnoise(1)
                  << std::endl;

        std::cout << "Time: " << time() << " ns" << std::endl;
        std::cout << "Signal chan(0) << " << signal(0) << " chan(1) " << signal(1) << std::endl;
        std::cout << "Clusters: " << std::endl;

        AsciiRoot::HitList::iterator ip;
        for (ip=begin(); ip!=end(); ++ip)
        {
            std::cout << "   chan: " << ip->center()
                      << " sig: "
                      << std::setw(6) << std::setprecision(1) << ip->signal()
                      << " left: " << ip->left() << " right: " << ip->right()
                      << std::endl;
            std::cout << '\t' << "channels: " << std::endl;
            int j;
            for (j=ip->left();j<=ip->right();j++)
                std::cout << "\t   " << j << " sn: " << _sn[j] << " signal: " << _signal[j] << " noise: " << _noise[j] << '\n';
            std::cout << std::endl;
        }

        cnvs->Update();
        ievt++;
    }
    std::cout << std::endl;
    _A_do_run= true;
    ::signal(SIGINT, old_handler);
}

bool is_text(const char *fnam)
{
    int nc;
    char buffer[1024];
    std::ifstream ifile(fnam);
    if (!fnam)
        return false;

    ifile.read(buffer, sizeof(buffer));
    nc = ifile.gcount();
    ifile.close();
    if (!nc) // empty files are text
    {
        return true;
    }

    std::string ss(buffer, nc);
    ifile.close();

    if ( ss.find('\0') != ss.npos )
        return false;

    double nontext = 0.;
    double ntotal = 0.;
    std::string::iterator ip;
    for (ip=ss.begin(); ip!=ss.end(); ++ip)
    {
        ntotal++;
        char c = *ip;
        if ( (c<' ' || c >'~') && !strchr("\n\t\r\b", c) )
            nontext++;
    }
    if ( nontext/ntotal > 0.3 )
        return false;

    return true;
}

