/*
 * TODO: may be cygwin can cope with this... Not sure.
 *       We stick, however, to MINGW
 *
 */
#include "USBserial.h"
#include "USBpatch.h"
#include <cstdio>
#include <cstring>
#include <cerrno>
#include <ctime>
#include <vector>
#include <glib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <GTimer.h>
#include "utils.h"

#ifndef G_OS_WIN32
#include <termios.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#else
#include "mingw.h"
#endif

#ifndef G_OS_WIN32
#ifndef FIONREAD
#define FIONREAD TIOCINQ
#endif
#endif



USBserial::USBserial() :
    fd(-1)
{

}

USBserial::USBserial(const char *fnam) :
    USBport(fnam), fd(-1)
{
    open(fnam);
}

USBserial::~USBserial()
{
}

/**
 * Opens the device
 */
int USBserial::open(const std::string &fnam)
{
#ifdef G_OS_UNIX

    int iok, ntry;

    // Abrimos el puerto
    devname = fnam;
    fd =:: open (devname.c_str(), O_RDWR | O_NOCTTY);
    if (fd < 0)
    {
        err_msg(_("Cannot open the device"));
        return -1;
    }


#ifdef __CYGWIN__
    speed_t speed=B115200;
#else    
    speed_t speed=B115200;// B115200;//B230400; //B1000000;//B9600;
#endif
    int csize=8;
    char partype='N';
    int stop=1;
    struct termios tconf;


    // cancels O_NDELAY
    iok = fcntl (fd, F_GETFL);
    //    fcntl (fd, F_SETFL, iok & ~O_NDELAY);


    // Get the attributes
    for (ntry = 0; ntry < 60; ntry++)
    {
        if (!(iok = tcgetattr (fd, &tconf)))
        {
            break;
        }
        g_usleep(1);
    }
    if (iok)
    {
        err_msg(_("Cannot get port attributes"));
        close();
        return -1;
    }
    // lo dejamos primero en un estado razonable
#ifndef __CYGWIN__    
//    cfmakeraw (&tconf);
#endif

    // Set in raw mode
    tconf.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL | IXON);
    tconf.c_oflag &= ~OPOST;
    tconf.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
    tconf.c_cflag &= ~(CSIZE | PARENB);
    
    // Set timeouts
    tconf.c_cc[VMIN]  = 1;
    tconf.c_cc[VTIME] = 0;

    // stop bits
    switch (stop)
    {
        case 2:
        case 1:
            tconf.c_cflag |= CSTOPB;
            break;
        default:
            break;
    }

    // data bits
    switch (csize)
    {
        case 5:
            tconf.c_cflag |= CS5;
            break;
        case 6:
            tconf.c_cflag |= CS6;
            break;
        case 7:
            tconf.c_cflag |= CS7;
            break;
        case 8:
            tconf.c_cflag |= CS8;
            break;
        default:
            err_msg (_("Invalid character size: %d\n"), csize);
            csize = 0;
    }

    // parity
    switch (partype)
    {
        case 'N':
        case 'n':
            tconf.c_cflag &= ~PARENB;
            break;
        case 'O':
        case 'o':
            tconf.c_cflag |= PARENB;
            tconf.c_cflag |= PARODD;
            break;
        case 'E':
        case 'e':
            tconf.c_cflag |= PARENB;
            tconf.c_cflag &= ~PARODD;
            break;
        case 's':
        case 'S':
            tconf.c_cflag &= ~PARENB;
            break;
        default:
            partype = 'K';
            err_msg (_("Invalid parity type\n"));
    }
    // speed
    cfsetispeed (&tconf, speed);
    cfsetospeed (&tconf, speed);

    iok = tcsetattr (fd, TCSANOW, &tconf);
#endif
    return 0;
}

int USBserial::close()
{
    return:: close(fd);
}

/**
 * Waits for data
 * */
int USBserial::wait_for_data(double tmout)
{
    int iok = poll(int(tmout*1000));
    if (iok)
        return 0;
    else
        return data_available(0);
}

unsigned int USBserial::data_available(unsigned int)
{
#ifdef G_OS_WIN32
    // TODO: See how we implement this in windows
    return -1;
#else
    int nread=0;
    ioctl(fd, FIONREAD, &nread);
    return nread;
#endif
}


int USBserial::read(void *buf, int size)
{

    return read_from_file(fd, buf, size);
}

int USBserial::write(void *buf, int size)
{
    int rc =:: write(fd, buf, size);
    return rc;
}

int USBserial::flush()
{
	char c;
	int sz, cnt=0;
	while ( (sz= data_available(0)) >0 )
	{
		read(&c, sizeof(char));
		cnt++;
	}
	return cnt;
}

int USBserial::poll(int timeout)
{
#ifdef G_OS_UNIX
    struct pollfd PS = { fd, POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL, 0 };

    if (fd<0)
    {
        printf("USBserial::poll - device not open\n");
        return -1;
    }

    int iok =:: poll(&PS, 1, timeout);
    {
        if ( PS.revents & POLLHUP )
        {
            err_msg(_("POLLHUP: Cable disconnected?"));
        }
        if ( PS.revents & POLLERR )
        {
            err_msg(_("POLLERR"));
        }
        if ( PS.revents & POLLNVAL )
        {
            err_msg(_("POLLNVAL"));
        }
    }
    if ( iok==1 && (PS.revents & POLLIN) )
    {
        return 0;
    }
    else
    {
        if (iok<0)
        {
            err_msg(_("error polling for data"));
            return iok;
        }
        else
        {
            return 1;
        }
    }
#endif
}
